package com.bergerkiller.bukkit.nolagg;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.bukkit.Bukkit;
import org.bukkit.ChatColor;
import org.bukkit.Chunk;
import org.bukkit.Material;
import org.bukkit.World;
import org.bukkit.command.Command;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Entity;
import org.bukkit.entity.ExperienceOrb;
import org.bukkit.entity.Item;
import org.bukkit.entity.Player;
import org.bukkit.entity.TNTPrimed;
import org.bukkit.event.Event;
import org.bukkit.event.Event.Priority;
import org.bukkit.plugin.Plugin;
import org.bukkit.plugin.PluginDescriptionFile;
import org.bukkit.plugin.PluginManager;
import org.bukkit.plugin.java.JavaPlugin;

import com.all.bukkit.threadcheck.ThreadCheck;
import com.bergerkiller.bukkit.config.ConfigurationNode;
import com.bergerkiller.bukkit.config.FileConfiguration;
import com.bergerkiller.bukkit.nolagg.ChunkOperation.Type;
import com.bergerkiller.bukkit.nolagg.meta.ChunkMetaData;
import com.bergerkiller.bukkit.nolagg.meta.WorldMetaData;
import com.bergerkiller.bukkit.nolagg.spawnlimit.SpawnHandler;

public class NoLagg extends JavaPlugin {
	public static NoLagg plugin;

	private final NLPlayerListener playerListener = new NLPlayerListener();
	private final NLBlockListener blockListener = new NLBlockListener();
	private final NLEntityListener entityListener = new NLEntityListener();
	private final NLWorldListener worldListener = new NLWorldListener();
	private final NLUnloadListener unloadListener = new NLUnloadListener();
	private int updateID = -1;
	private int ticktaskID = -1;
	public int updateInterval;
	
	public static boolean bufferItems;
	public static boolean bufferTNT;
	public static boolean useSpawnLimits;
	public static boolean isShowcaseEnabled = false;
	public static boolean isSCSEnabled = false;
	public static Plugin bleedingMobsInstance = null;
	
	private static Logger logger = Logger.getLogger("Minecraft");
	public static void log(Level level, String message) {
		logger.log(level, "[NoLagg] " + message);
	}
	
	public void onEnable() {
		ThreadCheck.init(this);
		
		plugin = this;
		
		//General registering
		PluginManager pm = getServer().getPluginManager();
		pm.registerEvent(Event.Type.PLAYER_PICKUP_ITEM, playerListener, Priority.Monitor, this);
		pm.registerEvent(Event.Type.PLAYER_CHANGED_WORLD, playerListener, Priority.Monitor, this);
		pm.registerEvent(Event.Type.PLAYER_JOIN, playerListener, Priority.Monitor, this);
		pm.registerEvent(Event.Type.CHUNK_LOAD, worldListener, Priority.Normal, this);
		pm.registerEvent(Event.Type.CHUNK_UNLOAD, worldListener, Priority.Lowest, this);
		pm.registerEvent(Event.Type.CHUNK_POPULATED, worldListener, Priority.Lowest, this);
		pm.registerEvent(Event.Type.LEAVES_DECAY, blockListener, Priority.Highest, this);
		pm.registerEvent(Event.Type.BLOCK_BREAK, blockListener, Priority.Highest, this);
		pm.registerEvent(Event.Type.BLOCK_PLACE, blockListener, Priority.Highest, this);
		pm.registerEvent(Event.Type.WORLD_LOAD, worldListener, Priority.Monitor, this);
		pm.registerEvent(Event.Type.WORLD_UNLOAD, worldListener, Priority.Monitor, this);
		pm.registerEvent(Event.Type.ITEM_SPAWN, entityListener, Priority.Normal, this);
		pm.registerEvent(Event.Type.CREATURE_SPAWN, entityListener, Priority.Lowest, this);
		//Make sure our events fires last, we don't want plugins reading the event after changes!
		getServer().getScheduler().scheduleSyncDelayedTask(this, new Runnable() {
			public void run() {
				PluginManager pm = getServer().getPluginManager();
				pm.registerEvent(Event.Type.ENTITY_EXPLODE, entityListener, Priority.Monitor, NoLagg.plugin);
				pm.registerEvent(Event.Type.ENTITY_DEATH, entityListener, Priority.Monitor, NoLagg.plugin);
				pm.registerEvent(Event.Type.CHUNK_UNLOAD, unloadListener, Priority.Monitor, NoLagg.plugin);
				isShowcaseEnabled = pm.isPluginEnabled("Showcase");
				isSCSEnabled = pm.isPluginEnabled("ShowCaseStandalone");
				bleedingMobsInstance = pm.getPlugin("BleedingMobs");
				if (ChunkSendQueue.enabled && pm.isPluginEnabled("Chunk Manager")) {
					log(Level.WARNING, "Chunk Manager detected, internal chunk sending system has been disabled.");
					log(Level.WARNING, "Either disable Chunk Manager or set chunkSending.enabled to false in the configuration.");
					ChunkSendQueue.enabled = false;
				}
				
			}
		}, 1);
		this.loadConfig();
		StackFormer.init();
		WorldMetaData.init();
		TnTHandler.init();
		PerformanceMonitor.init();
		updateID = getServer().getScheduler().scheduleSyncRepeatingTask(this, new Runnable() {
			public void run() {
				try {				
					WorldMetaData.updateAll();
				} catch (Throwable t) {
					log(Level.SEVERE, "An error occured while performing a routine update:");
					t.printStackTrace();
				}
			}
		}, 0, updateInterval);
		ticktaskID = getServer().getScheduler().scheduleSyncRepeatingTask(this, new Runnable() {
			public void run() {
				StopWatch.start();
				ChunkSendQueue.updateAll();
				PerformanceMonitor.chunkSend.update();
			}
		}, 1, 1);
		
		ChunkSaving.init();
		ChunkScheduler.init();
		AsyncAutoSave.init();
		AutoSaveChanger.init();
		ChunkSendQueue.init();
		BufferedPacket51MapChunk.init();
		
		getCommand("nolagg").setExecutor(this);
		
        //final msg
        PluginDescriptionFile pdfFile = this.getDescription();
        log(Level.INFO, pdfFile.getName() + " version " + pdfFile.getVersion() + " is enabled!" );
	}
	public void onDisable() {
		ThreadCheck.deinit();
		getServer().getScheduler().cancelTask(updateID);
		getServer().getScheduler().cancelTask(ticktaskID);
		TnTHandler.deinit();
		PerformanceMonitor.deinit();
		ChunkScheduler.deinit();
		AsyncAutoSave.deinit();
		ChunkSendQueue.deinit();
		SpawnHandler.clear();
		AutoSaveChanger.deinit();
		ChunkSaving.deinit();
		WorldMetaData.deinit();
		StackFormer.deinit();
		BufferedPacket51MapChunk.deinit();
		bleedingMobsInstance = null;
		log(Level.INFO, "NoLagg disabled!");
	}
	
	public void loadConfig() {
		//General settings
		FileConfiguration config = new FileConfiguration(this);
		config.load();
		bufferItems = config.get("bufferItems", true);
		bufferTNT = config.get("bufferTNT", true);
		useSpawnLimits = config.get("useSpawnLimits", true);
		ChunkMetaData.useUnloadDelay = config.get("useChunkUnloadDelay", true);
		ChunkMetaData.unloadDelay = config.get("chunkUnloadDelay", 10000);
		TnTHandler.interval = config.get("tntDetonationInterval", 1);
		TnTHandler.rate = config.get("tntDetonationRate", 10);
		TnTHandler.explosionRate = config.get("explosionRate", 5);
		ItemHandler.maxItemsPerChunk = config.get("maxItemsPerChunk", 40);
		ItemHandler.formStacks = config.get("formItemStacks", true);
		AutoSaveChanger.newInterval = config.get("autoSaveInterval", 400);
		updateInterval = config.get("updateInterval", 20);
		StackFormer.stackRadiusSquared = config.get("stackRadius", 1.0);
		StackFormer.stackRadiusSquared *= StackFormer.stackRadiusSquared;
		StackFormer.stackThreshold = config.get("stackThreshold", 2);
		PerformanceMonitor.monitorInterval = config.get("monitorInterval", 40);
		ChunkSendQueue.enabled = config.get("chunkSending.enabled", true);
		ChunkSendQueue.minRate = config.get("chunkSending.minRate", 0.25);
		ChunkSendQueue.maxRate = config.get("chunkSending.maxRate", 2.00);
		ChunkSendQueue.initialBoost = config.get("chunkSending.boost", 25);
		SpawnHandler.clear();
		if (useSpawnLimits) {
			//Spawn restrictions
			ConfigurationNode limits = config.getNode("spawnlimits");
			ConfigurationNode node;
			node = limits.getNode("default");
			for (String key : node.getKeys()) {
				SpawnHandler.WORLDDEFAULT.setLimit(key, node.get(key, -1));
			}
			node = limits.getNode("global");
			for (String key : node.getKeys()) {
				SpawnHandler.GLOBAL.setLimit(key, node.get(key, -1));
			}
			node = limits.getNode("worlds");
			for (ConfigurationNode cn : node.getNodes()) {
				String world = cn.getName();
				for (String key : cn.getKeys()) {
					SpawnHandler.getWorldLimiter(world).setLimit(key, node.get(key, -1));
				}
			}
		}
		ItemHandler.clearIgnoredTypes();
		for (String type : config.getList("ignoredItemTypes", String.class)) {
			Material mat = Material.matchMaterial(type);
			if (mat != null) {
				ItemHandler.addIgnoredType(mat);
			} else {
				log(Level.WARNING, "Unknown item type found in configuration: " + type);
			}
		}

		config.trim();
		//Write out data
		config.save(); 
	}
			
	private String[] lastargs = new String[1];
	
	@Override
	public boolean onCommand(CommandSender sender, Command cmd, String cmdLabel, String[] args) {
		if (args.length > 0) {
			boolean all = args[0].equalsIgnoreCase("clearall");
			if (args[0].equalsIgnoreCase("clear") || all) {
				if (sender instanceof Player) {
					if (!((Player) sender).hasPermission("nolagg.clear")) {
						sender.sendMessage(ChatColor.DARK_RED + "You don't have permission to use this command!");
						return true;
					}
				} else {
					all = true;
				}
				//fix and partly read args
				boolean tnt = args.length == 1;
				boolean items = tnt;
				boolean animals = false;
				boolean monsters = false;
				boolean remall = false;
				boolean last = args.length == 2 && args[1].equalsIgnoreCase("last");
				if (last) args = lastargs;
				String[] toremove = new String[args.length - 1];
				for (int i = 1; i < args.length; i++) {
					String name = args[i].toLowerCase();
					if (name.contains("tnt")) {
						tnt = true;
					} else if (name.contains("item")) {
						items = true;
					} else if (name.contains("animal")) {
						animals = true;
					} else if (name.contains("monster")) {
						monsters = true;
					} else if (name.contains("mob")) {
						animals = true;
						monsters = true;
					} else if (name.contains("all")) {
						remall = true;
					}
					toremove[i - 1] = name;
				}
				args = toremove;
				World[] worlds;
				if (all) {
					worlds = Bukkit.getServer().getWorlds().toArray(new World[0]);
					if (items) WorldMetaData.clearAllItems();
				} else {
					worlds = new World[] {((Player) sender).getWorld()};
					if (items) WorldMetaData.get(worlds[0]).clearItems();
				}
				if (tnt) TnTHandler.clear();
				int remcount = 0;
				for (World world : worlds) {
					for (Entity e : world.getEntities()) {
						boolean remove = false;
						if (e instanceof Player) {
							continue;
						} else if (remall) {
							remove = true;
					    } else if (args.length == 0) {
							remove = e instanceof Item || e instanceof TNTPrimed || e instanceof ExperienceOrb;
						} else if (e instanceof TNTPrimed && tnt) {
							remove = true;
						} else if (e instanceof Item && items) {
							remove = true;
						} else {
							String type = e.getClass().getSimpleName().toLowerCase();
							if (e instanceof Item) {
								type = "item" + ((Item) e).getItemStack().getType().toString().toLowerCase();
							} else if (type.startsWith("craft")) {
								type = type.substring(5);
							} else if (type.contains("tnt")) {
								type = "tnt";
							}
							if (animals && SpawnHandler.isAnimal(type)) {
								remove = true;
							} else if (monsters && SpawnHandler.isMonster(type)) {
								remove = true;
							} else {
								for (String arg : args) {
									if (type.contains(arg) || arg.contains(type)) {
										remove = true;
										break;
									}
								}	
							}
						}
						if (remove) {
							e.remove();
							remcount++;
						}
					}
				}
				if (last) {
					sender.sendMessage(ChatColor.GREEN + "The last-used clear command has been invoked:");
				}
				if (all) {
					sender.sendMessage(ChatColor.YELLOW + "All worlds have been cleared: " + remcount + " entities removed!");
				} else {
					sender.sendMessage(ChatColor.YELLOW + "This world has been cleared: " + remcount + " entities removed!");
				}
			} else if (args[0].equalsIgnoreCase("monitor")) {
				if (sender instanceof Player) {
					Player p = (Player) sender;
					if (!p.hasPermission("nolagg.monitor")) {
						sender.sendMessage(ChatColor.DARK_RED + "You don't have permission to use this command!");
						return true;
					}
					if (PerformanceMonitor.recipients.remove(p.getName())) {
						for (int i = 0; i < 10; i++) {
							p.sendMessage(" ");
						}
						p.sendMessage(ChatColor.YELLOW + "You are no longer monitoring this server.");
					} else {
						PerformanceMonitor.recipients.add(p.getName());
						p.sendMessage(ChatColor.GREEN + "You are now monitoring this server.");
						p.sendMessage(ChatColor.GREEN + "To stop monitoring, perform this command again.");
					}
				} else {
					if (PerformanceMonitor.sendConsole) {
						PerformanceMonitor.sendConsole = false;
						sender.sendMessage("You are no longer monitoring this server.");
					} else {
						PerformanceMonitor.sendConsole = true;
						sender.sendMessage("You are now monitoring this server.");
						sender.sendMessage("To stop monitoring, perform this command again.");
					}
				}
			} else if (args[0].equalsIgnoreCase("log")) {
				if (sender instanceof Player) {
					if (!((Player) sender).hasPermission("nolagg.log")) {
						sender.sendMessage(ChatColor.DARK_RED + "You don't have permission to use this command!");
						return true;
					}
				}
				if (PerformanceMonitor.sendLog) {
					PerformanceMonitor.sendLog = false;
					sender.sendMessage("Server stats are no longer logged to file.");
				} else {
					PerformanceMonitor.sendLog = true;
					sender.sendMessage("Server stats are now logged to file.");
					sender.sendMessage("To stop logging, perform this command again.");
				}
			} else if (args[0].equalsIgnoreCase("clearlog")) {
				if (sender instanceof Player) {
					if (!((Player) sender).hasPermission("nolagg.clearlog")) {
						sender.sendMessage(ChatColor.DARK_RED + "You don't have permission to use this command!");
						return true;
					}
				}
				if (PerformanceMonitor.clearLog()) {
					sender.sendMessage("Server log cleared");
				} else {
					sender.sendMessage("Failed to clear the server log");
				}
			} else if (args[0].equalsIgnoreCase("reload")) {
				if (sender instanceof Player) {
					if (!((Player) sender).hasPermission("nolagg.reload")) {
						sender.sendMessage(ChatColor.DARK_RED + "You don't have permission to use this command!");
						return true;
					}
				}
				this.loadConfig();
				sender.sendMessage("Configuration of NoLagg reloaded");
			} else if (args[0].equalsIgnoreCase("fix")) {
				if (sender instanceof Player) {
					Player p = (Player) sender;
					if (!p.hasPermission("nolagg.fix")) {
						sender.sendMessage(ChatColor.DARK_RED + "You don't have permission to use this command!");
						return true;
					}
					int radius = Bukkit.getServer().getViewDistance();
					if (args.length == 2) {
						try {
							radius = Integer.parseInt(args[1]);
						} catch (Exception ex) {}
					}
					int cx = p.getLocation().getBlockX() >> 4;
					int cz = p.getLocation().getBlockZ() >> 4;
					for (int a = -radius; a <= radius; a++) {
						for (int b = -radius; b <= radius; b++) {
							Chunk c = p.getWorld().getChunkAt(cx + a, cz + b);
							ChunkScheduler.schedule(c, Type.LIGHTING);
						} 
					}
					p.sendMessage(ChatColor.GREEN + "A " + (radius * 2 + 1) + " X " + (radius * 2 + 1) + " chunk area around you is currently being fixed from lighting issues...");
				} else {
					sender.sendMessage("This command is only for players!");
				}
			} else {
				sender.sendMessage("Unknown sub-command!");
			}
		} else {
			sender.sendMessage("NoLagg version " + this.getDescription().getVersion() + ", for available commands see the the Bukkit(Dev) page of NoLagg.");
		}
		return true;
	}
	
}
